(    Title:  Root module tests
    Author:  David N. Williams
      File:  root-module-test.fs
   License:  Public Domain
   Version:  0.8.1
   Revised:  November 12, 18, 2010
   Revised:  December 19, 21, 2010
   Revised:  October 5, 2011

This version includes tests for Bruce McFarling's helper words,
which are in root-module.fs 0.8.1, but not the more minimal
0.8.2.
)
decimal

\ for debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]
: .#       ( n -- )  cr ." #" . cr ;

s" ttester-xf.fs"     INCLUDED
s" otester.fs"        INCLUDED
s" tester-display.fs" INCLUDED
s" xtester-errors.fs" INCLUDED
s" otester-errors.fs" INCLUDED

true VERBOSE !

\ before first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

: ?.errors  ( -- )
  VERBOSE @
  IF blue-text ." XT-#ERRORS: " normal-text XT-#ERRORS @ .
     3 spaces
     blue-text ." OT-#ERRORS: " normal-text OT-#ERRORS @ .
  THEN ;

: ={  O{ T{ ;    : =>   O-> -> ;    : }=  }O }T ;

\ for debugging
: .#order  ( n -- )  cr 2 .r ." : " get-order .s drops ;

PFE-HOST [IF]    ?." Host is pfe, "
[THEN]
GFORTH-HOST [IF] ?." Host is gforth, "
[THEN]
IFORTH-HOST [IF] ?." Host is iForth, "
[THEN]
?." root-module version is 0.8.1.


\ *** ORDER STACK UTILITIES
(
We don't want to load order.fs, because then the conditional
copies of order stack definitions in root-module.fs would not be
used.
)

: o@  ( o: wid -- ) ( -- wid )
(
A version of order@ for use before root-module is loaded.
)
  get-current definitions get-current swap set-current ;

[UNDEFINED] -ROT [IF] : -ROT  rot rot ; [THEN]

: _-ORDER ( w n*wid n -- n'*wid w n' )                \ order.fs
(
A factor in -ORDER.  Assume that n is nonnegative.  The input
wid sequence n*wid may or may not contain copies of w.  The
output n'*wid has any copies removed, and n' is n minus the
number removed.  The implementation uses the return stack to
invert the order of wids, and recursively removes them to build
the output on the data stack.
)
  dup IF
    1- swap >r  ( >> ) RECURSE ( << )  ( i*wid w i) ( R: wid')
    over r@ xor IF 1+ r> -rot EXIT THEN r> drop
  THEN ;

: -ORDER  ( wid -- )  ( o: i*x -- j*x )               \ order.fs
(
Remove all copies of the wid from the order stack.  "minus-order"
)
  get-order _-ORDER nip set-order ;


\ *** TESTS

TESTING  root module state

get-current CONSTANT env-current
         o@ CONSTANT env-context

\ this test does not leave MODULES in the search order
={ s" root-module.fs" INCLUDED => MODULES }=
T{ get-current o@ -> env-current env-context }T

TESTING  ORDER-DEPTH BEGIN-MODULE:

={ get-order => get-order drop order-depth }=

0 VALUE private-wid   0 VALUE public-wid

={ BEGIN-MODULE: Foo order> to private-wid order> to public-wid => }=

T{ get-current -> private-wid }T

\ wids not in the original order
={ public-wid  -ORDER => }=
={ private-wid -ORDER => }=

={ MODULES Foo order> => MODULES public-wid }=

TESTING  PRIVATE: PUBLIC:
(
Only the compilation word list changes, not the search order.
)

={ PRIVATE: get-current => private-wid }=
={ PUBLIC: get-current => public-wid }=

TESTING  END-MODULE

\ set state just after BEGIN-MODULE:
T{ PRIVATE: public-wid >order private-wid >order -> }T

\ drops wids
={ END-MODULE public-wid >order private-wid >order => }=

\ restores current
T{ get-current -> env-current }T

\ back to original state
T{ END-MODULE get-current order@ -> env-current env-context }T

TESTING  WORDLIST: SET-HELPER HELPER:

0 VALUE Bar-wid

={ WORDLIST: Bar => }=
={ Bar order-depth order@ to Bar-wid
     => order-depth 1+ Bar-wid >order }=
={ Bar set-helper => }=
={ HELPER: get-current => Bar-wid }=

?.errors GFORTH-HOST [IF] ?.cr [THEN]
