(      Title:  Public and private word support for modules
        File:  root-module.fs
   Test File:  root-module-test.fs
  Adapted by:  David N. Williams
     License:  Public Domain
     Version:  0.8.1
     Revised:  December 22, 2010

This library uses design features from the Forth Scientific
Library fsl-util auxiliary files, Neal Bridges' modules.fs,
experimental work by Bruce McFarling, and an experimental
attempt by the author.  See the comp.lang.forth discussion:

  http://tinyurl.com/pubpriv-clf
)


\ *** GENERAL USE

[UNDEFINED] drops [IF]                 \ BMcF
: drops ( +n -- ) 0 ?DO drop LOOP ; [THEN]

[UNDEFINED] order-drops [IF]
: order-drops  ( +n -- ) ( o: wid_n ... wid_1 -- )
  0 ?DO previous LOOP ; [THEN]

[UNDEFINED] order-depth [IF]
: order-depth  ( -- n )  get-order dup >r drops r> ; [THEN]

[UNDEFINED] >order [IF]
: >order  ( wid -- order: wid )
  >r get-order r> swap 1+ set-order ; [THEN]

[UNDEFINED] order> [IF]
: order>  ( o: wid -- ) ( -- wid )
  get-order swap >r 1- set-order r> ; [THEN]

[UNDEFINED] order@ [IF]
: order@  ( o: wid -- ) ( -- wid )
  get-order over >r drops r> ; [THEN]

[UNDEFINED] WORDLIST: [IF]              \ BMcF, NBr
\ define a conventional vocabulary word
: WORDLIST:  ( "name" -- )   wordlist CREATE ,
DOES>  ( o: -- wid )  @ >order ; [THEN]

\ *** PRIVATE DEFINITIONS

order-depth get-current wordlist
( private) dup set-current >order

( odepth comp.wid)
VARIABLE initial-defs         ( comp.wid) initial-defs !
VARIABLE initial-order-depth    ( odepth) initial-order-depth !
VARIABLE private-defs         get-current private-defs !
VARIABLE public-defs       initial-defs @ public-defs !

\ *** PUBLIC DEFINITIONS

public-defs @ set-current  \ same as initial-defs

: PRIVATE:  ( -- )  private-defs @ set-current ;
: PUBLIC:   ( -- )  public-defs  @ set-current ; 

\ BEGIN-MODULE: and END-MODULE are not to be nested.

WORDLIST: MODULES

: BEGIN-MODULE:  ( "name" -- ) ( o: -- public private )
(
Save information that allows END-MODULE to restore the initial
search order and compilation wordlist, assuming the initial
order has remained unchanged.  Define "name" as a vocabulary
word belonging to the MODULES wordlist, and add its wordlist,
"public", to the search order.  Create an anonymous wordlist,
"private", add it to the search order, and make it the
compilation wordlist.  Store the public and private wids for use
by PUBLIC: and PRIVATE:.
)
  get-current initial-defs !
  modules definitions previous
  wordlist CREATE dup , dup >order public-defs !
  wordlist dup >order private-defs !
  definitions    \ default seection is private
DOES> ( o: -- public ) @ >order ;

: END-MODULE  ( o: <extras> -- )
(
Restore the initial search order and compilation wordlist.  An
ambiguous condition exists if the initial search order has been
clobbered.
)
  initial-defs @ set-current
  order-depth initial-order-depth @ - order-drops ;

\ *** HELPER WORDS
(
This section is a variation on Bruce McFarling's version.

Note that helpers are ordinary vocabularies, not modules.  The
following syntax patterns may be used inside or outside a
module:

  WORDLIST: <name>   \ define a named vocabulary
  <name> set-helper  \ make <name> the vocabulary for
                     \ definitions in subsequent helper sections,
                     \ without adding it to the search order
  <name>             \ add the <name> vocabulary's wid to the search order

Note that HELPER: can in principle be used outside a normal
module, whereas PUBLIC: and PRIVATE: currently cannot.
)

PRIVATE:
VARIABLE helper-defs   0 helper-defs !

PUBLIC:
: set-helper  ( o: wid -- )  order> helper-defs ! ;
: HELPER:     ( -- )         helper-defs @ set-current ;

END-MODULE
MODULES  \ add to the search order


0 [IF]  \ EXAMPLE LAYOUT
(
Assume that root-module.fs is loaded, and hence that MODULES is
in the search order.
)

\ Load unloaded module, helper, and other files.

WORDLIST: <new_helper>    \ define a new helper vocabulary word
                          \ which is not to belong to a module
                          \ wordlist
<helper_1> ... <helper_m> \ add helper vocabularies to the search order

<helper> order> set-current  \ make a helper vocabulary the
                             \ compilation wordlist without
                             \ adding it to the search order
\ OR
<helper> set-helper
HELPER:

\ Add modules to the search order which are to remain visible
\ after END-MODULE.  Pre-module code like this is legal, but we
\ think not best practise, because putting the list of modules
\ on which it depends inside a module more clearly documents its
\ dependencies, and adding modules outside makes it easier to
\ introduce inadvertent, unwanted dependencies.
\
<mod_1> ... <mod_n>

BEGIN-MODULE: Foo  
\ Any modules or helpers added to the search order after this
\ point are dropped by END-MODULE.

\ Load unloaded module helper, or other files.

<mod_n+1> ...    \ add modules to the search order
<helper_m+1> ... \ add helpers to the search order

<initial private Foo definitions>

PUBLIC:
\ additional public definitions

WORDLIST: <new_helper>  \ define a new helper vocabulary word
                        \ belonging to this module's public wordlist

<helper> set-helper     \ make a helper the compilation vocabulary
                        \ for the next helper sections

<helper>                \ add a helper to the search order

\ Any number of private, public, and helper sections in any order:

PRIVATE:
\ additional private definitions

PUBLIC:
\ additional public definitions

HELPER:
\ additional helper definitions

<another_helper> set-helper  \ use <another_helper> in the next
                             \ helper sections
\ etc.
END-MODULE
[THEN]  \ END EXAMPLE LAYOUT
